from __future__ import division, absolute_import, print_function, unicode_literals
from awlsim.common.compat import *


VERSION_MAJOR = 0
VERSION_MINOR = 36

VERSION_STRING = "%d.%d" % (VERSION_MAJOR, VERSION_MINOR)
